/*
	g++ -o pbm.anova pbm.anova.cpp random.cpp
*/

#include <cstdlib>
#include <iostream>
#include <cmath>
#include <list>
#include <vector>
#include <map>
#include <string>
#include <algorithm>
#include <unistd.h>
using namespace std;

#include "random.h"

#define square(x) (double)((double)(x)*(x))
#define LINESIZE 1024

static KW_RNG::RNG _my_random;
void my_srandom(int x) { _my_random.init(x,!x*13,x*x+1,(x>>16)+(x<<16)); }
double my_random01() { return _my_random.rand_closed01(); }
double my_uniform(double x) {return _my_random.uniform(0.0,x); }
double my_range(double x, double y) {return _my_random.uniform(x,y); }
int my_binomial(double pp, int n) { return _my_random.binomial(pp,n); }
double my_normal(double mu = 0.0, double sd = 1.0) { return _my_random.normal(mu,sd); }
double my_chisq(double df) { return _my_random.chi_square(df); }
double my_expo(double lambda) { return _my_random.exponential(lambda); }
double my_beta(double a1, double a2) { return _my_random.beta(a1, a2); }

#ifdef _WIN32
#include <process.h>
#include <windows.h>
void set_priority_low() {
  HANDLE hProcess=OpenProcess(PROCESS_ALL_ACCESS,TRUE,_getpid());
  SetPriorityClass(hProcess,IDLE_PRIORITY_CLASS);
}
#else
#include <unistd.h>
#endif

char inputDir[LINESIZE];
char outputDir[LINESIZE];
char fileName[LINESIZE];

const unsigned int nK = 32896; // number of k-mers

unsigned int nT = 0; // number of TFs
unsigned int nF = 0; // number of families
unsigned int nSample = 100; // number of mcmc samples
unsigned int nBurn = 100; // number of burn-in samples
double logPriorOddsFam = -6.0; // priors for family specific
double logPriorOddsTf = -6.0; // priors for tf specific
double nu0 = 1.0, sigma0 = 1.0;
double mu0 = 3.0, limitFam = 1.0;
double theta0 = 3.0, limitTf = 1.0;

double meanAll = 0.0; // overall mean
double sdAll = 1.0; // overall standard deviation

double kappa = 1.0; // variance scale for family specific effect
double lambda = 1.0; // variance scale for k-mer background effect
double sigma = 1.0; // standard deviation of residuals
double *theta1; // mean of tf specific binding nT
double *theta2; // mean of tf specific non-binding nT
double *mu1; // mean of family specific binding nF
double *mu2; // mean of family specific non-binding nF
double *tau; // mean of k-mer background
double *tfSum; // sum of tf specific binding socre
double *tfNoSum; // sum of tf specific non-binding socre
unsigned int *tfNum; // number of tf specific binding
unsigned int *tfNoNum; // number of tf specific non-binding
double *famSum; // sum of family specific binding score
double *famNoSum; // sum of family specific non-binding score
unsigned int *famNum; // number of family specific binding
unsigned int *famNoNum; // number of family specific non-binding

double **matScore; // matrix of E-scores nK x nT
double **familyBindProb; // family binding probability from model nK x nF
double **familyBindFreq; // family binding frequency from sampling nK x nF
double **familyNoBindProb; // family non-binding probability from model nK x nF
double **familyNoBindFreq; // family non-binding frequency from sampling nK x nF
double **tfBindProb; // TF-specific binding probability from model nK x nT
double **tfBindFreq; // TF-specific binding frequency from sampling nK x nT
double **tfNoBindProb; // TF-specific non-binding probability from model nK x nT
double **tfNoBindFreq; // TF-specific non-binding frequency from sampling nK x nT

unsigned int **familyBind; // family binding indicator nK x nF
unsigned int **familyNoBind; // family non-binding indicator nK x nF
unsigned int **tfBind; // TF-specific binding indicator nK x nT
unsigned int **tfNoBind; // TF-specific non-binding indicator nK x nT
double **familySum; // family sum of E-scores nK x nT
double **familySqr; // family sum of square nK x nT
double *nullSum; // effective		sum of E-score in null group for each k-mer nK
double *nullSize; // effective number of TFs in null group of each k-mer nK

unsigned int *familyId; // family ID of TFs nT
unsigned int *familySize; // number of TFs in each familiy nF

/**
  *External function
  */
int getopt(int, char *const *, const char *);

/**
  *Take integer power of an integer a^n
  */
unsigned int power(unsigned int a, unsigned int n) 
{
    if(n == 0) 
		return 1;
	else if(n == 1)
		return a;
	else {
		unsigned int b = a * a;
		return n&1 ? a * power( b, n>>1 ) : power(b, n>>1);
	}
}

/**
 * Function to measure time elapsed
 */
double diffClock(clock_t clock1,clock_t clock2)
{
	double diffticks = clock1-clock2;
	double diffms = (diffticks*1000)/CLOCKS_PER_SEC;
	return diffms;
} 

/**
 * Error message
 */
void errorExit(const char* msg){
	fprintf(stderr, "%s\n", msg);
	exit(0);
}

void initMemory()
{
	unsigned int i,j,k;

    familySum = (double **) calloc(nK, sizeof(double*));
	familySqr = (double **) calloc(nK, sizeof(double*));
    familyBindProb = (double **) calloc(nK, sizeof(double*));
	familyBindFreq = (double **) calloc(nK, sizeof(double*));
	familyNoBindProb = (double **) calloc(nK, sizeof(double*));
	familyNoBindFreq = (double **) calloc(nK, sizeof(double*));
	familyBind = (unsigned int **) calloc(nK, sizeof(unsigned int*));
	familyNoBind = (unsigned int **) calloc(nK, sizeof(unsigned int*));
	tfBindProb = (double **) calloc(nK, sizeof(double*));
	tfBindFreq = (double **) calloc(nK, sizeof(double*));
	tfNoBindProb = (double **) calloc(nK, sizeof(double*));
	tfNoBindFreq = (double **) calloc(nK, sizeof(double*));
	tfBind = (unsigned int **) calloc(nK, sizeof(unsigned int*));
	tfNoBind = (unsigned int **) calloc(nK, sizeof(unsigned int*));

	theta1 = (double *) calloc(nT, sizeof(double));
	theta2 = (double *) calloc(nT, sizeof(double));
    mu1 = (double*) calloc(nF, sizeof(double));
	mu2 = (double*) calloc(nF, sizeof(double));
	tau = (double*) calloc(nK, sizeof(double));
	tfSum = (double*) calloc(nT, sizeof(double));
	tfNoSum = (double*) calloc(nT, sizeof(double));
	tfNum = (unsigned int *) calloc(nT, sizeof(unsigned int));
	tfNoNum = (unsigned int *) calloc(nT, sizeof(unsigned int));
	famSum = (double*) calloc(nF, sizeof(double));
	famNoSum = (double*) calloc(nF, sizeof(double));
	famNum = (unsigned int *) calloc(nF, sizeof(unsigned int));
	famNoNum = (unsigned int *) calloc(nF, sizeof(unsigned int));
	nullSum = (double*) calloc(nK, sizeof(double));
	nullSize = (double *) calloc(nK, sizeof(double));
	
	for( i = 0; i < nT; i++)
	{
		theta1[i] = 0.0;
		theta2[i] = 0.0;
		tfSum[i] = 0.0;
		tfNum[i] = 0;
		tfNoSum[i] = 0.0;
		tfNoNum[i] = 0;
	}
	for( j = 0; j < nF; j++)
	{
		mu1[j] = 0.0;
		mu2[j] = 0.0;
		famSum[j] = 0.0;
		famNum[j] = 0;
		famNoSum[j] = 0.0;
		famNoNum[j] = 0;
	}

	for( k = 0; k < nK; k++)
	{
		tfBindProb[k] = (double *) calloc(nT, sizeof(double));
		tfNoBindProb[k] = (double *) calloc(nT, sizeof(double));
		tfBindFreq[k] = (double *) calloc(nT, sizeof(double));
		tfNoBindFreq[k] = (double *) calloc(nT, sizeof(double));
		tfBind[k] = (unsigned int *) calloc(nT, sizeof(unsigned int));
		tfNoBind[k] = (unsigned int *) calloc(nT, sizeof(unsigned int));

		familyBindProb[k] = (double *) calloc(nF, sizeof(double));
		familyBindFreq[k] = (double *) calloc(nF, sizeof(double));
		familyNoBindProb[k] = (double *) calloc(nF, sizeof(double));
		familyNoBindFreq[k] = (double *) calloc(nF, sizeof(double));
		familySum[k] = (double *) calloc(nF, sizeof(double));
		familySqr[k] = (double *) calloc(nF, sizeof(double));
		familyBind[k] = (unsigned int *) calloc(nF, sizeof(unsigned int));
		familyNoBind[k] = (unsigned int *) calloc(nF, sizeof(unsigned int));

		nullSum[k] = 0.0;
		nullSize[k] = 0.0;
		tau[k] = 0.0;
		for( i = 0; i < nT; i++)
		{	
			tfBindProb[k][i] = 0.0;
			tfBindFreq[k][i] = 0.0;
			tfBind[k][i] = 0;
			tfNoBindProb[k][i] = 0.0;
			tfNoBindFreq[k][i] = 0.0;
			tfNoBind[k][i] = 0;
		}

		for( j = 0; j < nF; j++)
		{	
			familySum[k][j] = 0.0;
			familySqr[k][j] = 0.0;
			familyBindProb[k][j] = 0.0;
			familyBindFreq[k][j] = 0.0;
			familyBind[k][j] = 0;
			familyNoBindProb[k][j] = 0.0;
			familyNoBindFreq[k][j] = 0.0;
			familyNoBind[k][j] = 0;
		}
	}
}

void readData()
{
	unsigned int i,j,k;
	
	float score = 0.0;
	int temp = 0;
	double sum=0.0, ssq=0.0;
	FILE *scoreFile, *familyFile;
	strcpy(fileName, inputDir);
	strcat(fileName,"/score.txt");
	scoreFile = fopen (fileName, "r");
	strcpy(fileName, inputDir);
	strcat(fileName,"/family.txt");
	familyFile = fopen (fileName, "r");

	if (scoreFile == NULL) perror ("Error opening score file");
	else if (familyFile == NULL) perror ("Error opening family file");
    else {
		fscanf (familyFile, "%d", &temp);
		nT = temp;
		fscanf (familyFile, "%d", &temp);
		nF = temp;
		cout << nT << "\t" << nF << endl;

		familyId = (unsigned int *) calloc(nT, sizeof(unsigned int));
		for( i = 0; i < nT; i++)
			familyId[i] = 0;
		familySize = (unsigned int *) calloc(nF, sizeof(unsigned int));
		for( j = 0; j < nF; j++)
			familySize[j] = 0;
		matScore =  (double **) calloc(nK, sizeof(double*));
		for(k = 0; k < nK; k++)
		{
			matScore[k] = (double *) calloc(nT, sizeof(double));
			for(i = 0; i < nT; i++)
				matScore[k][i] = 0.0;
		}

		for( k = 0; k < nK; k++)
		{	
			for( i = 0; i < nT; i++)
			{
				fscanf (scoreFile, "%f", &score);
				matScore[k][i] = score;
				sum += score;
				ssq += square(score);
			}
		}
		
		for( i = 0; i < nT; i++)
		{
			fscanf (familyFile, "%d", &temp);
			familyId[i] = temp;
			familySize[temp] += 1;
		}
    }
	meanAll = sum/(double)(nK*nT);
	ssq = ssq/(double)(nK*nT) - square(meanAll);
	sdAll = sqrt(ssq);
    cout << meanAll << "\t" << sdAll << endl;
	fclose(scoreFile);
	fclose(familyFile);
}

void initFamilyBind(double cutoff)
{
	unsigned int i,j,k;
	double score = 0.0, mean=0.0, variance=0.0;
	for( k = 0; k < nK; k++)
	{
		mean = 0.0;
		for( i = 0; i < nT; i++) 
		{
			j = familyId[i];
			score = matScore[k][i];
			score -= meanAll;
			score /= sdAll;
			matScore[k][i] = score;
			familySum[k][j] += score;
			familySqr[k][j] += square(score);
			// do not use large families to estimate background
			mean += score/familySize[j];
		}
		mean /= (double)nF;
		tau[k] = mean;	
		
		for( j = 0; j < nF; j++)
		{
			mean = familySum[k][j]/(double)familySize[j]-tau[k];
			variance = familySqr[k][j] - familySize[j]*square(mean);
			if(familySize[j] > 1) 
				variance /= ((double)familySize[j] - 1.0); 
			variance /= (double)familySize[j];
			if( mean > cutoff*sqrt(square(sigma)+variance) )
			{
				familyBind[k][j] = 1;
				famSum[j] += mean;
				famNum[j] += 1;
			} else if( mean < -cutoff*sqrt(square(sigma)+variance) )
			{
				familyNoBind[k][j] = 1;
				famNoSum[j] += mean;
				famNoNum[j] += 1;
			}
		}
	}
	
	for( j = 0; j < nF; j++)
	{
		if(famNum[j] > 0)
			mu1[j] = famSum[j]/(double)famNum[j];
		else
			mu1[j] = cutoff;

		if(famNoNum[j] > 0)
			mu2[j] = famNoSum[j]/(double)famNoNum[j];
		else
			mu2[j] = -cutoff;
	}
}

void initTfBind(double cutoff)
{
	unsigned int i,j,k;
	double score = 0.0;

	for( k = 0; k < nK; k++)
	{
		for( i = 0; i < nT; i++) 
		{
			j = familyId[i];
			score = matScore[k][i]-familyBind[k][j]*mu1[j]-familyNoBind[k][j]*mu2[j]-tau[k];
			if(score >= cutoff) 
			{
				tfBind[k][i] = 1;
				tfSum[i] += score;
				tfNum[i] += 1;
			} else if(score <= -cutoff) 
			{
				tfNoBind[k][i] = 1;
				tfNoSum[i] += score;
				tfNoNum[i] += 1;
			}
		}
	}
	for( i = 0; i < nT; i++)
	{
		if (tfNum[i] > 0)
			theta1[i] = tfSum[i]/(double)tfNum[i];
		else
			theta1[i] = cutoff;

		if (tfNoNum[i] > 0)
			theta2[i] = tfNoSum[i]/(double)tfNoNum[i];
		else
			theta2[i] = -cutoff;
	}
}

//Given familyBind, mu, theta, and sigma, update tfBind
void updateTfBind(unsigned int update)
{
	unsigned int i,j,k;
	double logPostOdds=0.0, logPostNoOdds=0.0, logLikDiff=0.0;
	double score=0.0, r=0.0, n=0.0, cum0, cum1, cum2;

	for( i = 0; i < nT; i++) 
	{
		tfSum[i] = 0.0;
		tfNum[i] = 0;
		tfNoSum[i] = 0.0;
		tfNoNum[i] = 0;
	}
	for(k = 0; k < nK; k++)
	{
		for(i = 0; i < nT; i++) 
		{
			j = familyId[i];
			n = kappa/(kappa+1.0);
			score = matScore[k][i]-familyBind[k][j]*mu1[j]-familyNoBind[k][j]*mu2[j]-tau[k];
			if(tfBind[k][j] == 1 && tfNoBind[k][j] ==1)
				errorExit("Conflict indicators for tf binding!");
			logPostOdds = logPriorOddsTf;
			logLikDiff = square(score)-square(score-theta1[i]);
			logPostOdds += ((0.5*n)/square(sigma))*logLikDiff;

			logPostNoOdds = logPriorOddsTf;
			logLikDiff = square(score)-square(score-theta2[i]);
			logPostNoOdds += ((0.5*n)/square(sigma))*logLikDiff;

			r = my_uniform(1.0);
			cum0 = exp(logPostOdds)+exp(logPostNoOdds)+1.0;
			cum1 = exp(logPostOdds)/cum0;
			cum2 = cum1 + exp(logPostNoOdds)/cum0;
			if(r < cum1 && score > 0.0)
			//if(r < cum1)
			{
				tfBind[k][i] = 1;
				tfNoBind[k][i] = 0;
				tfSum[i] += score;
				tfNum[i] += 1;
			} else if(r < cum2 && score < 0.0)
			//} else if(r < cum2)
			{
				tfNoBind[k][i] = 1;
				tfBind[k][i] = 0;
				tfNoSum[i] += score;
				tfNoNum[i] += 1;
			} else {
				tfBind[k][i] = 0;
				tfNoBind[k][i] = 0;
			}

			if(update == 1)
			{
				tfBindFreq[k][i] += tfBind[k][i];
				tfNoBindFreq[k][i] += tfNoBind[k][i];
			}
		}	
	}
}

void regenerateSufficientStat()
{
	unsigned int i,j,k;
	double score=0.0, temp=0.0;
	for(k = 0; k < nK; k++)
	{
		nullSum[k] = 0.0;
		nullSize[k] = 0.0;
		for(j = 0; j < nF; j++)
		{
			familySum[k][j] = 0.0;
			familySqr[k][j] = 0.0;
		}
		for(i = 0; i < nT; i++) 
		{
			j = familyId[i];
			score = matScore[k][i] - tfBind[k][i]*theta1[i] - tfNoBind[k][i]*theta2[i] - tau[k];
			familySum[k][j] += score;
			familySqr[k][j] += square(score);
		}
	}
}

//Given tfBind, mu, theta, and sigma, update familyBind
void updateFamilyBind(unsigned int update) 
{
	double score=0.0, r=0.0, temp=0.0, mean=0.0, logPostOdds=0.0, logPostNoOdds=0.0, logLikDiff=0.0;
	double cum0, cum1, cum2, n;
	unsigned int i,j,k;
	for(j = 0; j < nF; j++)
	{
		famSum[j] = 0.0;
		famNum[j] = 0;
		famNoSum[j] = 0.0;
		famNoNum[j] = 0;
	}
	for(k = 0; k < nK; k++) 
	{
		nullSum[k] = 0.0;
		nullSize[k] = 0.0;
		for(j = 0; j < nF; j++)
		{
			logPostOdds = logPriorOddsFam;
			logLikDiff = 2.0*familySum[k][j]*mu1[j] - familySize[j]*square(mu1[j]);
			logLikDiff = kappa*logLikDiff/(double)(kappa+familySize[j]);
			logPostOdds += (0.5/square(sigma))*logLikDiff;
			
			logPostNoOdds = logPriorOddsFam;
            logLikDiff = 2.0*familySum[k][j]*mu2[j] - familySize[j]*square(mu2[j]);
			logLikDiff = kappa*logLikDiff/(double)(kappa+familySize[j]);
            logPostNoOdds += (0.5/square(sigma))*logLikDiff;

			r = my_uniform(1.0);
			cum0 = exp(logPostOdds)+exp(logPostNoOdds)+1.0;
			cum1 = exp(logPostOdds)/cum0;
			cum2 = cum1 + exp(logPostNoOdds)/cum0;
			n = kappa/(double)(familySize[j]+kappa);
			if(r < cum1 && familySum[k][j] > 0.0) 
			//if(r < cum1)
			{
				familyBind[k][j] = 1;
				familyNoBind[k][j] = 0;
				double mean0 = nullSum[k]/nullSize[k];
				nullSum[k] += n*(familySum[k][j]+familySize[j]*tau[k]-familySize[j]*mu1[j]);
				nullSize[k] += n*familySize[j];
				famSum[j] += familySum[k][j]/(double)familySize[j];
				famNum[j] += 1;
				//cout << familySum[k][j]/(double)familySize[j]-tau[k] << "\t" << logPostOdds << "\t" << logPostNoOdds << "\t" << cum1 << "\t" << cum2 << "\t" << r << endl;	
			} else if(r < cum2 && familySum[k][j] < 0.0)
			//} else if(r < cum2)
			{
				familyNoBind[k][j] = 1;
				familyBind[k][j] = 0;
				double mean0 = nullSum[k]/nullSize[k];
				nullSum[k] += n*(familySum[k][j]+familySize[j]*tau[k]-familySize[j]*mu2[j]);
				nullSize[k] += n*familySize[j];
				famNoSum[j] += familySum[k][j]/(double)familySize[j];
				famNoNum[j] += 1;
			} else
			{
				familyBind[k][j] = 0;
				familyNoBind[k][j] = 0;
				nullSum[k] += n*(familySum[k][j]+familySize[j]*tau[k]);
				nullSize[k] += n*familySize[j];
			} 

			if(update == 1)
			{
				familyBindFreq[k][j] += familyBind[k][j];
				familyNoBindFreq[k][j] += familyNoBind[k][j];
			}
		}
	}
}

void updateTfMean(double theta0, double limitTf)
{
	double n0 = 1.0, n;
	unsigned int i,j;
	double mean, sd, temp;
	for(i = 0; i < nT; i++)
	{
		j = familyId[i];
		n = kappa/(kappa+1.0);
		mean = (n*tfSum[i]+n0*theta0)/(double)(n*tfNum[i]+n0);
		sd = sigma/sqrt(n*tfNum[i]+n0);
		temp = my_normal(mean,sd);
		if(theta1[i] < limitTf)
			theta1[i] = limitTf;
		if(temp > limitTf)
			theta1[i] = temp;
		else
			theta1[i] = limitTf;

		//cout << "Tf "<< i << ":\t" << mean << "\t" << tfNum[i] << "\t" << theta1[i] << endl;
		
		mean = (n*tfNoSum[i]-n0*theta0)/(double)(n*tfNoNum[i]+n0);
		sd = sigma/sqrt(n*tfNoNum[i]+n0);
		temp = my_normal(mean,sd);
		if(theta2[i] > -limitTf)
			theta2[i] = -limitTf;
		if(temp < -limitTf)
			theta2[i] = temp;
		else
			theta2[i] = -limitTf;

		//cout << "Tf "<< i << ":\t" << mean << "\t" << tfNoNum[i] << "\t" << theta2[i] << endl;
	}
}

void updateFamilyMean(double mu0, double limitFam)
{
	double n0 = 1.0, n;
	double mean, sd, temp;
	for(unsigned int j = 0; j < nF; j++)
	{
		n = kappa*familySize[j]/(kappa+familySize[j]);
		mean = (n*famSum[j]+n0*mu0)/(double)(n*famNum[j]+n0);
		sd = sigma/sqrt(n*famNum[j]+n0);
		temp = my_normal(mean,sd);
		if(mu1[j] < limitFam)
			mu1[j] = limitFam;
		if(temp > limitFam)
			mu1[j] = temp;
		else
			mu1[j] = limitFam;
		
		//cout << "Family "<< j << ":\t" << mean << "\t" << temp << "\t" << mu1[j] << "\t" << famNum[j] << "\t" << familySize[j] << endl;

		mean = (n*famNoSum[j]-n0*mu0)/(double)(n*famNoNum[j]+n0);
		sd = sigma/sqrt(n*famNoNum[j]+n0);
		temp = my_normal(mean,sd);
		if(mu2[j] > -limitFam)
			mu2[j] = -limitFam;
		if(temp < -limitFam)
			mu2[j] = temp;
		else
			mu2[j] = -limitFam;
		
		//cout << "Family "<< j << ":\t" << mean << "\t" << temp << "\t" < mu2[j] << "\t" << famNum[j] << "\t" << familySize[j] << endl;
	}
}

void updateKmersMean()
{
	double mean, sd, temp, avg;
	double n0=1.0, kappa0=1.0,  nu0=1.0, lambda0=1.0; 
	double ssq=nu0/kappa0, nu=nu0, ssq2=nu0/lambda0, nu2=nu0;
	for(unsigned int k = 0; k < nK; k++)
	{
		mean = nullSum[k]/(nullSize[k]+n0);
		sd = sigma/(lambda*sqrt(nullSize[k]+n0));
		tau[k] = my_normal(mean,sd);
        ssq2 += square(tau[k]);
        nu2 += 1;
		for(unsigned int j = 0; j < nF; j++)
		{
			if(familyBind[k][j] == 1 && familyNoBind[k][j] ==1)
				errorExit("Conflict indicators for family binding!");
			if(familyBind[k][j] == 1)
			{
				mean = (familySum[k][j]+kappa*mu1[j])/(familySize[j]+kappa);
				sd = sigma/sqrt(familySize[j]+kappa);
				temp = my_normal(mean,sd);
				ssq += square(temp-mu1[j]);
				nu += 1;
			} else if(familyNoBind[k][j] == 1)
			{
				mean = (familySum[k][j]+kappa*mu2[j])/(familySize[j]+kappa);
				sd = sigma/sqrt(familySize[j]+kappa);
				temp = my_normal(mean,sd);
				ssq += square(temp-mu2[j]);
				nu += 1;
			} else {
				mean = familySum[k][j]/(double)(familySize[j]+kappa);
				sd = sigma/sqrt(familySize[j]+kappa);
				temp = my_normal(mean,sd);
				ssq += square(temp);
				nu += 1;
			}
		}

	}
	kappa = square(sigma)*my_chisq(nu)/ssq;
    lambda = square(sigma)*my_chisq(nu2)/ssq2;
}

// include theta's and mu's priors?
double updateSigma(double sigma0, double nu0)
{
	unsigned int i,j,k;
	double n0=1.0, llk = 0.0;
	double ssq = nu0*square(sigma0), nu = nu0+nK*(nT+1);
	for(k = 0; k < nK; k++)
	{
		for(j = 0; j < nF; j++)
		{	
			if(familyBind[k][j] == 1 && familyNoBind[k][j] ==1)
				errorExit("Conflict indicators for family binding!");
			if(familyBind[k][j] == 1)
			{
				ssq += familySqr[k][j]-2.0*familySum[k][j]*mu1[j]+familySize[j]*square(mu1[j]);
				ssq -= square(familySum[k][j]-familySize[j]*mu1[j])/(double)(familySize[j]+kappa);
				llk += 0.5*(log(kappa)-log(familySize[j]+kappa));
			} else if(familyNoBind[k][j] == 1)
			{
				ssq += familySqr[k][j]-2.0*familySum[k][j]*mu2[j]+familySize[j]*square(mu2[j]);
				ssq -= square(familySum[k][j]-familySize[j]*mu2[j])/(double)(familySize[j]+kappa);
				llk += 0.5*(log(kappa)-log(familySize[j]+kappa)); 
			} else {
				ssq += familySqr[k][j];
				ssq -= square(familySum[k][j])/(double)(familySize[j]+kappa);
				llk += 0.5*(log(kappa)-log(familySize[j]+kappa)); 
			}
		}
        ssq += n0*square(tau[k])/lambda;
	}
	sigma = sqrt(ssq/my_chisq(nu));
	llk -= ssq/(2.0*square(sigma));
	llk -= nK*(nT+1)*log(sigma);

	//cout << "sigma:\t" << sigma << endl;
	
	return(llk);
}

void updateBindProb()
{
	double score=0.0, r=0.0, cum0=0.0, n=0.0;
	double logPostOdds=0.0, logPostNoOdds=0.0, logLikDiff=0.0;
	unsigned int i,j,k;
	for(k = 0; k < nK; k++) 
	{
		nullSum[k] = 0.0;
		nullSize[k] = 0.0;
		for(j = 0; j < nF; j++)
		{
            logPostOdds = logPriorOddsFam;
			logLikDiff = 2.0*familySum[k][j]*mu1[j] - familySize[j]*square(mu1[j]);
			logLikDiff = kappa*logLikDiff/(double)(kappa+familySize[j]);
			logPostOdds += (0.5/square(sigma))*logLikDiff;
			
			logPostNoOdds = logPriorOddsFam;
            logLikDiff = 2.0*familySum[k][j]*mu2[j] - familySize[j]*square(mu2[j]);
			logLikDiff = kappa*logLikDiff/(double)(kappa+familySize[j]);
            logPostNoOdds += (0.5/square(sigma))*logLikDiff;

			cum0 = exp(logPostNoOdds)+1.0;
			familyBindProb[k][j] = logPostOdds - log(cum0);
			cum0 = exp(logPostOdds)+1.0;
			familyNoBindProb[k][j] = logPostNoOdds - log(cum0);
		}
		for( i = 0; i < nT; i++) 
		{

			j = familyId[i];
			score = matScore[k][i]-familyBind[k][j]*mu1[j]-familyNoBind[k][j]*mu2[j]-tau[k];
			n = kappa/(kappa+1.0);
			logPostOdds = logPriorOddsTf;
			logLikDiff = square(score)-square(score-theta1[i]);
			logPostOdds += ((0.5*n)/square(sigma))*logLikDiff;

			logPostNoOdds = logPriorOddsTf;
			logLikDiff = square(score)-square(score-theta2[i]);
			logPostNoOdds += ((0.5*n)/square(sigma))*logLikDiff;

			cum0 = exp(logPostNoOdds)+1.0;
			tfBindProb[k][i] = logPostOdds - log(cum0);
			cum0 = exp(logPostOdds)+1.0;
			tfNoBindProb[k][i] = logPostNoOdds - log(cum0);
		}	
	}
}

void writeResults()
{
	FILE *tfBindFreqFile, *tfNoBindFreqFile, *familyBindFreqFile, *familyNoBindFreqFile;
	FILE *tfBindProbFile, *tfNoBindProbFile, *familyBindProbFile, *familyNoBindProbFile;

	strcpy(fileName, outputDir); strcat(fileName,"/tfbindfreq.txt"); tfBindFreqFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/tfnobindfreq.txt"); tfNoBindFreqFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/familybindfreq.txt"); familyBindFreqFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/familynobindfreq.txt"); familyNoBindFreqFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/tfbindprob.txt"); tfBindProbFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/tfnobindprob.txt"); tfNoBindProbFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/familybindprob.txt"); familyBindProbFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/familynobindprob.txt"); familyNoBindProbFile = fopen (fileName, "w");

	unsigned int i,j,k;
	for(k = 0; k < nK; k++)
	{
		for(j = 0; j < nF; j++)
		{
			familyBindFreq[k][j]=(double)(familyBindFreq[k][j])/(double)(nSample);
			fprintf(familyBindFreqFile,"%f\t",familyBindFreq[k][j]);
			familyNoBindFreq[k][j]=(double)(familyNoBindFreq[k][j])/(double)(nSample);
			fprintf(familyNoBindFreqFile,"%f\t",familyNoBindFreq[k][j]);
			fprintf(familyBindProbFile,"%f\t",familyBindProb[k][j]);
			fprintf(familyNoBindProbFile,"%f\t",familyNoBindProb[k][j]);
		}
		fprintf(familyBindFreqFile,"\n");
		fprintf(familyNoBindFreqFile,"\n");
		fprintf(familyBindProbFile,"\n");
		fprintf(familyNoBindProbFile,"\n");
		for(i = 0; i < nT; i++)
		{
			tfBindFreq[k][i]=(double)(tfBindFreq[k][i])/(double)(nSample);
			tfNoBindFreq[k][i]=(double)(tfNoBindFreq[k][i])/(double)(nSample);
			fprintf(tfBindFreqFile,"%.2f\t",tfBindFreq[k][i]);
			fprintf(tfBindProbFile,"%.2f\t",tfBindProb[k][i]);
			fprintf(tfNoBindFreqFile,"%.2f\t",tfNoBindFreq[k][i]);
			fprintf(tfNoBindProbFile,"%.2f\t",tfNoBindProb[k][i]);
		}
		fprintf(tfBindFreqFile,"\n");
		fprintf(tfBindProbFile,"\n");
		fprintf(tfNoBindFreqFile,"\n");
		fprintf(tfNoBindProbFile,"\n");
	}

	fclose(tfBindFreqFile);
	fclose(tfNoBindFreqFile);
	fclose(familyBindFreqFile);
	fclose(familyNoBindFreqFile);
	fclose(tfBindProbFile);
	fclose(tfNoBindProbFile);
	fclose(familyBindProbFile);
	fclose(familyNoBindProbFile);
}

void cleanUp()
{
	for(unsigned int k = 0; k < nK; k++)
	{
		free(matScore[k]);
		free(tfBindProb[k]);
		free(tfBindFreq[k]);
		free(tfBind[k]);
		free(familyBindProb[k]);
		free(familyBindFreq[k]);
		free(familyBind[k]);
		free(familyNoBindProb[k]);
		free(familyNoBindFreq[k]);
		free(familyNoBind[k]);
		free(familySum[k]);
		free(familySqr[k]);
	}
	free(matScore);
	free(tfBindProb);
	free(tfBindFreq);
	free(tfBind);
	free(tfNoBindProb);
	free(tfNoBindFreq);
	free(tfNoBind);
	free(familyBindProb);
	free(familyBindFreq);
	free(familyBind);
	free(familyNoBindProb);
	free(familyNoBindFreq);
	free(familyNoBind);
	
	free(familySum);
	free(familySqr);
	free(familyId);
	free(familySize);
	free(nullSum);
	free(nullSize);
	free(theta1);
	free(theta2);
	free(mu1);
	free(mu2);
	free(tau);
	free(tfSum);
	free(tfNoSum);
	free(tfNum);
	free(tfNoNum);
	free(famSum);
	free(famNoSum);
	free(famNum);
	free(famNoNum);
}

void printUsage()
{
	fprintf(stderr, "pbm.anova usage:\n\n");
	fprintf(stderr, "\t-i <input directory, containing family.txt, score.txt>\n");
	fprintf(stderr, "\t-o <output directory>\n");
	fprintf(stderr, "\t-m <number of burn-in iterations, default: 100>\n");
	fprintf(stderr, "\t-n <number of mcmc sampling iterations, default: 100>\n");
	fprintf(stderr, "\t-p <log prior for family specific binding, default: -6.0>\n");
	fprintf(stderr, "\t-q <log prior for tf specific binding, default: -6.0>\n");
	exit(0);
}

void parseInput(int argc, char *argv[])
{
	extern char *optarg;
	float tempFloat;
	int opt, tempInt;

	while ((opt = getopt(argc, argv, "i:o:m:n:p:q:"))!=EOF)
	{
		switch(opt)
		{
			case 'i':
				sscanf(optarg, "%s", inputDir);
				break;
			case 'o':
				sscanf(optarg, "%s", outputDir);
				break;
			case 'm':
				if(sscanf(optarg, "%d", &tempInt) && tempInt > 0)
					nBurn = tempInt;
				break;
			case 'n':
				if(sscanf(optarg, "%d", &tempInt) && tempInt > 0)
					nSample = tempInt;
				break;
			case 'p':
				if(sscanf(optarg, "%f", &tempFloat))
					logPriorOddsFam = tempFloat;
				break;
			case 'q':
				if(sscanf(optarg, "%f", &tempFloat))
					logPriorOddsTf = tempFloat;
				break;
		}
	}	
}

int main(int argc, char *argv[]) {

#ifdef _WIN32
	clock_t __clock = clock();
	time_t __time;
	set_priority_low();
	my_srandom(__clock + time(&__time) + _getpid());
#else
	my_srandom(getpid());
#endif

	if (argc == 1)
		printUsage();

	clock_t begin, end;
	unsigned int i,j,k;
	double cutoffTf = 2.0, cutoffFam = 2.0;

	parseInput(argc, argv);
	readData();
	initMemory();
	initFamilyBind(cutoffFam);
	initTfBind(cutoffTf);

	strcpy(fileName, outputDir); strcat(fileName,"/param.txt"); FILE *paramFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/trace.txt"); FILE *traceFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/omega1.txt"); FILE *mu1File = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/omega2.txt"); FILE *mu2File = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/delta1.txt"); FILE *theta1File = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/delta2.txt"); FILE *theta2File = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/tau.txt"); FILE *tauFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/tfbindnum.txt"); FILE *tfBindNumFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/tfnobindnum.txt"); FILE *tfNoBindNumFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/familybindnum.txt"); FILE *familyBindNumFile = fopen (fileName, "w");
	strcpy(fileName, outputDir); strcat(fileName,"/familynobindnum.txt"); FILE *familyNoBindNumFile = fopen (fileName, "w");

	fprintf(paramFile,"nT\tnF\tnK\tnSample\tnBurn\tlogPriorOddsFam\tlogPriorOddsTf\tmu0\tlimitFam\ttheta0\tlimitTf\tsigma0\tnu0\n");
	fprintf(paramFile,"%d\t%d\t%d\t%d\t%d\t%f\t%f\t%f\t%f\t%f\t%f\t%f\t%f\n",nT,nF,nK,nSample,nBurn,logPriorOddsFam,logPriorOddsTf,mu0,limitFam,theta0,limitTf,sigma0,nu0);
	fprintf(traceFile,"gamma1\tgamma2\tsigma\tlogLik\n");
	begin=clock();
	double* logLik = (double *) calloc(nSample+nBurn, sizeof(double));
	double* tauMean = (double *) calloc(nK, sizeof(double));
	for(unsigned int iter = 0; iter < nSample+nBurn; iter++) 
	{
		logLik[iter] = 0.0;
		if(iter >= nBurn)
			updateTfBind(1);
		else
			updateTfBind(0);
		updateTfMean(theta0, limitTf);
		regenerateSufficientStat();
		if(iter >= nBurn)
			updateFamilyBind (1); 
		else
			updateFamilyBind (0); 
		updateFamilyMean(mu0, limitFam);
        logLik[iter] = updateSigma(sigma0, nu0);
        updateKmersMean();
		fprintf(traceFile,"%f\t%f\t%f\t%f\n",lambda,kappa,sigma,logLik[iter]);
		
		if(iter >= nBurn) 
		{
			for(j = 0; j < nF; j++)
			{
				fprintf(familyBindNumFile,"%d\t",famNum[j]);
				fprintf(familyNoBindNumFile,"%d\t",famNoNum[j]);
				if(famNum[j] > 0)
					fprintf(mu1File,"%f\t",mu1[j]);
				else 
					fprintf(mu1File,"0\t");
				if(famNoNum[j] > 0)
					fprintf(mu2File,"%f\t",mu2[j]);
				else
					fprintf(mu2File,"0\t");
			}
			fprintf(mu1File,"\n");
			fprintf(mu2File,"\n");
			fprintf(familyBindNumFile,"\n");
			fprintf(familyNoBindNumFile,"\n");
			for(i = 0; i < nT; i++)
			{
				fprintf(tfBindNumFile,"%d\t",tfNum[i]);
				fprintf(tfNoBindNumFile,"%d\t",tfNoNum[i]);
				if(tfNum[j] > 0)
					fprintf(theta1File,"%f\t",theta1[i]);
				else
					fprintf(theta1File,"0\t");
				if(tfNoNum[j] > 0)
					fprintf(theta2File,"%f\t",theta2[i]);
				else
					fprintf(theta2File,"0\t");
			}
			fprintf(theta1File,"\n");
			fprintf(theta2File,"\n");
			fprintf(tfBindNumFile,"\n");
			fprintf(tfNoBindNumFile,"\n");
			for(k=0; k < nK; k++)
				tauMean[k] += tau[k];
			/*
			for(k=0; k < nK; k++)
				fprintf(tauFile,"%f\t",tau[k]);
			fprintf(tauFile,"\n");
			*/
		}
		cout << "Iteratoin: " << iter+1 << "\tgamma1: " << lambda  << "\tgamma2: " << kappa << "\tsigma: " << sigma << "\tLog-Likelihood: " << logLik[iter] << endl;
    }
	for(k=0; k < nK; k++)
		fprintf(tauFile,"%f\n",tauMean[k]/(double)nSample);
	end = clock();
	cout << "Total running time: " << diffClock(end,begin) << "ms" << endl;
	cout << "Running time per iteration: " << diffClock(end,begin)/(double)(nBurn+nSample) << "ms" << endl;

	regenerateSufficientStat();
	updateBindProb();
	writeResults();
	cleanUp();
	fclose(paramFile);
	fclose(traceFile);
	fclose(mu1File);
	fclose(mu2File);
	fclose(theta1File);
	fclose(theta2File);
	fclose(tauFile);
	fclose(familyNoBindNumFile);
	fclose(familyBindNumFile);
	fclose(tfBindNumFile);
	fclose(tfNoBindNumFile);
	free(logLik);
	free(tauMean);

	return 0;
}
