/*
	gcc -lm -o pbm.motif pbm.motif.c
*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <sys/time.h>

#define MAXLINE 16383
#define ALPHASIZE 4
#define A 0
#define C 1
#define G 2
#define T 3
#define N 4
#define square(x) (double)((double)(x)*(x))

/* for random number generator */
#define IA 16807
#define IM 2147483647
#define AM (1.0 / IM)
#define IQ 127773
#define IR 2836

/* some ascii for keys */
#define ENDLINE '\0'
#define TAB 9
#define SPACE 32
#define TILTA 126

char fileName[MAXLINE];
int nL = 10; // length
int nE = 4; // ending non-gap minimum
int nB = 4; // beginning gap maximum
int nD = 1; // gap in init, (nL-8)/2
int nS = 0;
int nSample = 1000; // number of mcmc samples
int nBurn = 1000; // number of burn-in samples
double prior = 0.1;

struct sequence {
  int id;
  int *seqi, *rseqi, len;
  int *alignment;
  double *score, *rscore;
  struct sequence *next;
};

struct sequence *seqs = NULL;
double **motifMatrix;
int **countMatrix;
int *position;
long idum;  /* for random number generator */

/**
  *External function
  */
int getopt(int, char *const *, const char *);

/**
 * Error message
 */
void errorExit(const char* msg){
	fprintf(stderr, "%s\n", msg);
	exit(0);
}

/**
 * init rand
 */
void initRand() {
  struct timeval tv;
  long temp = 0;

  while (1) {
    gettimeofday(&tv, NULL);
    if (tv.tv_usec != temp) {
      idum = tv.tv_usec;
      return;
    }
  }
}

/*
 * Generate a double random number between 0 and 1. 
 */
double drand() {
  long k;
  
  k = idum / IQ;
  idum = IA * (idum - k * IQ) - IR * k;
  if (idum < 0) idum += IM;
  return (AM * idum);
}

int mystrlen(char *string) {
  int len = strlen(string);
  char *c;

  c = &(string[len-1]);

  /* everything outside SPACE and TILTA '~' does not have a viewable
   * character representation of it */
  while (*c < SPACE || *c > TILTA) {
    *c = ENDLINE;
    c--;
  }
  return (c - string + 1);
}

void seqatoi(char *line, int *seqi, int *rseqi, int len) {
  int i;

  for (i = 0; i < len; i++) {
    switch (line[i]) {
    case 'A': seqi[i] = A; rseqi[len-i-1] = T; break;
	case 'C': seqi[i] = C; rseqi[len-i-1] = G; break;
	case 'G': seqi[i] = G; rseqi[len-i-1] = C; break;
	case 'T': seqi[i] = T; rseqi[len-i-1] = A; break;
	}
  }
}

void seqitoa(char *str, int *seq, int len) {
    int i;
    
    for (i = 0; i < len; i++) {
        switch (seq[i]) {
            case A: str[i] = 'A'; break;
            case C: str[i] = 'C'; break;
            case G: str[i] = 'G'; break;
            case T: str[i] = 'T'; break;
            case N: str[i] = 'N'; break;
        }
    }
}


void createSeq(char *seq, int id, struct sequence **tail) {
  struct sequence *s = NULL;
  int len;
  
  s = (struct sequence *) malloc (sizeof (struct sequence));
  s->id = id;
  s->len = len = strlen(seq); 

  s->seqi = (int *) malloc (len * sizeof(int));
  s->score = (double *) malloc ((len+nB-nE+1) * sizeof(double));
  s->rseqi = (int *) malloc (len * sizeof(int));
  s->alignment = (int *) malloc (nL * sizeof(int));
  s->rscore = (double *) malloc ((len+nB-nE+1) * sizeof(double));
  seqatoi(seq, s->seqi, s->rseqi, len);
  s->next = NULL;
  if (seqs) {
	(*tail)->next = s;
    (*tail) = s;
  } else
    seqs = *tail = s;
}

void readSeqFile(FILE *fp) {
  char seq[MAXLINE], line[MAXLINE];
  int len, id;
  struct sequence *tail = seqs;

  id = 0;
  while (fgets(line, MAXLINE, fp)) {
    len = mystrlen(line);

    if (len == 0) 
      continue;
	strcpy(seq, line);
	createSeq(seq, id, &tail);
	id += 1;
  }
  nS = id;
  position = (int*) malloc (nS * sizeof(int));
  fclose(fp);
}

void updateCountMatrix()
{
	int i,j,start,end,pos,len,overlap;
	int *s, *rs;
	struct sequence *seq = seqs;
	double score=0.0, cum=0.0, temp=0.0, r=0.0;
	for(i = 0; i < nL; i++)
	{
		for(j = 0; j < 4; j++)
			countMatrix[i][j] = 0;
	}
	while(seq != NULL)
	{
		cum = 0.0;
		len = seq->len;
		s = seq->seqi;
		rs = seq->rseqi;
		for(i = -nB; i < len-nE+1; i++)
		{
			start = i<0 ? 0 : i;
			end = (i+nL)<len ? (i+nL) : len;
			score = 0.0;
			for(j = start; j < end; j++)
				score += motifMatrix[j-i][s[j]];
			score += -log(ALPHASIZE)*(nL+start-end);
			seq->score[nB+i] = score;
			cum += exp(score);
			score = 0.0;
			for(j = start; j < end; j++)
				score += motifMatrix[j-i][rs[j]];
			score += -log(ALPHASIZE)*(nL+start-end);
			seq->rscore[nB+i] = score;
			cum += exp(score);
		}	

		r = cum*drand();

		temp = 0.0;
		for(i = -nB; i < len-nE+1; i++)
		{
			temp += exp(seq->score[nB+i]);
			if(temp > r)
			{
				start = i<0 ? 0 : i;
				end = (i+nL)<len ? (i+nL) : len;
				score = 0.0;
				for(j = start; j < end; j++)
					countMatrix[j-i][s[j]] += 1;
				
				position[seq->id] = i+nB;
				break;
			}
			temp += exp(seq->rscore[nB+i]);
			if(temp > r)
			{
				start = i<0 ? 0 : i;
				end = (i+nL)<len ? (i+nL) : len;
				score = 0.0;
				for(j = start; j < end; j++)
					countMatrix[j-i][rs[j]] += 1;

				position[seq->id] = -(i+nB);
				break;
			}
		}
		seq = seq->next;
	}

}

void align()
{
	int i,j,len;
    int pos,start,end,ori;
	int *s, *rs;
	struct sequence *seq = seqs;
	double score=0.0;
	while(seq != NULL)
	{
		len = seq->len;
		s = seq->seqi;
		rs = seq->rseqi;
        score = -1024.0;
        pos = 0;
        start = 0;
        end = 0;
        ori = 0;
		for(i = -nB; i < len-nE+1; i++)
		{
            if(seq->score[nB+i] > score)
            {
                pos = i>0 ? 0 : -i;
                start = i<0 ? 0 : i;
                end = (i+nL)<len ? (i+nL) : len;
                ori = 0;
                score = seq->score[nB+i];
            } else if (seq->rscore[nB+i] > score) {
                pos = i>0 ? 0 : -i;
                start = i<0 ? 0 : i;
                end = (i+nL)<len ? (i+nL) : len;
                ori = 1;
                score = seq->score[nB+i]; 
            }
		}	
        for(j = 0; j < nL; j++)
        {
            seq->alignment[j] = N;
        }
        for(i = pos, j = start; j < end; i++,j++)
        {
            if(ori==0)
            {
                seq->alignment[i] = s[j];
            } else {
                seq->alignment[i] = rs[j];
            }
            
        }
        seq = seq->next;
    }
}

void updateMotifMatrix()
{
	int i,j;
	double sum = 0.0;
	for(i = 0; i < nL; i++)
	{
		sum = 4*prior;
		for(j = 0; j < 4; j++)
			sum += countMatrix[i][j];
		for(j = 0; j < 4; j++)
			motifMatrix[i][j] = log((countMatrix[i][j]+prior)/sum);
	}

}

void initCountMatrix()
{
	int i,j;
	for(i = 0; i < nL; i++)
	{
		for(j = 0; j < 4; j++)
			countMatrix[i][j] = 0;
	}
	for(i = 0; i < seqs->len; i++)
		countMatrix[nD+i][seqs->seqi[i]] += 0.1;
}

void printUsage()
{
	fprintf(stderr, "pbm.motif usage:\n\n");
	fprintf(stderr, "\t-i <input file>\n");
	fprintf(stderr, "\t-m <number of burn-in iterations, default: 100>\n");
	fprintf(stderr, "\t-n <number of mcmc sampling iterations, default: 100>\n");
	exit(0);
}

void parseInput(int argc, char *argv[])
{
	extern char *optarg;
	float tempFloat;
	int opt, tempInt;

	while ((opt = getopt(argc, argv, "i:m:n:"))!=EOF)
	{
		switch(opt)
		{
			case 'i':
				sscanf(optarg, "%s", fileName);
				break;
			case 'm':
				if(sscanf(optarg, "%d", &tempInt) && tempInt > 0)
					nBurn = tempInt;
				break;
			case 'n':
				if(sscanf(optarg, "%d", &tempInt) && tempInt > 0)
					nSample = tempInt;
				break;
		}
	}	
}

void cleanUp()
{
	int i;
	for(i = 0; i < nL; i++)
	{
		free(motifMatrix[i]);
		free(countMatrix[i]);
	}
	free(motifMatrix);
	free(countMatrix);
	free(position);
	
	struct sequence *seq = seqs, *tail;
	while(seq != NULL)
	{
		tail = seq->next;
		free(seq->seqi);
		free(seq->rseqi);
		free(seq->score);
		free(seq->rscore);
        free(seq->alignment);
		seq = tail;
	}
}

int main(int argc, char *argv[]) {

	if (argc == 1)
		printUsage();

	int i,j,t;
	initRand();
	parseInput(argc, argv);
	FILE *inputFile = fopen (fileName, "r");
	readSeqFile(inputFile);
	double ** resultMatrix = (double **) malloc (nL * sizeof(double*));
	motifMatrix = (double **) malloc (nL * sizeof(double*));
	countMatrix = (int **) malloc (nL * sizeof(int*));
	for(i = 0; i < nL; i++)
	{
		resultMatrix[i] = (double *) malloc (4 * sizeof(double));
		motifMatrix[i] = (double *) malloc (4 * sizeof(double));
		countMatrix[i] = (int *) malloc (4 * sizeof(int));
		for(j = 0; j < 4; j++)
		{
			motifMatrix[i][j] = 0.0;
			resultMatrix[i][j] = 0.0;
			countMatrix[i][j] = 0;
		}
	}

	initCountMatrix();
	updateMotifMatrix();
	
	/*
	for(i = 0; i < nL; i++)
	{
		for(j = 0; j < 4; j++)
			printf("%f\t",motifMatrix[i][j]);
		
		printf("\n");
	}
	*/
	
	for(t = 0; t < nBurn+nSample; t++)
	{
		updateCountMatrix();
		updateMotifMatrix();
		//printf("\n%d\n",t+1);
		if(t >= nBurn)
		{
			for(i = 0; i < nL; i++)
			{
				for(j = 0; j < 4; j++)
				{
					resultMatrix[i][j] += motifMatrix[i][j];
					//printf("%f\t",motifMatrix[i][j]);
				}
				//printf("\n");
			}
		}
	}

    align();
    FILE *alignFile = fopen ("seqalign.txt", "w");
    for(i = 0; i < nL; i++)
	{
        fprintf(alignFile,"%d",i);
    }
    fprintf(alignFile,"\n");
    int *s;
    char str[MAXLINE];
	struct sequence *seq = seqs;
	while(seq != NULL)
	{
        s = seq->alignment;
        seqitoa(str,s,nL);
        fprintf(alignFile,"%s\n",str);
        printf("%s\n",str);
        seq = seq->next;
    }    
	//for(i = 0; i < nS; i++)
	//	printf("%d\n",position[i]);

	strcat(fileName,".pwm"); 
    FILE *outputFile = fopen (fileName, "w");
	fprintf(outputFile,"A\tC\tG\tT\n");
	for(i = 0; i < nL; i++)
	{
		for(j = 0; j < 4; j++)
			fprintf(outputFile,"%f\t",resultMatrix[i][j]/(double)nSample);
		fprintf(outputFile,"\n");
	}
	cleanUp();
}


